#  shellyhtcollector, a python module to process sensor readings from Shelly H&T devices
#
# (C) 2022 Michel Anders (varkenvarken)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  version: 20220814111956

import argparse
from datetime import datetime
from sys import stderr, exit
from os import environ

from dateutil import tz

from .Server import Interceptor
from .Database import MeasurementDatabase


# all arguments/options can be set using environment variables or command line options
# if both are present, command line options take precedence


def get_args(arguments=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--database",
        type=str,
        default=environ.get("MARIADB_DATABASE", "shellyht"),
        help="database schema",
    )
    parser.add_argument(
        "--dbuser", type=str, default=environ.get("MARIADB_USER", "=test-user")
    )
    parser.add_argument(
        "--dbpassword", type=str, default=environ.get("MARIADB_PASSWORD", "test_secret")
    )
    parser.add_argument(
        "--dbhost",
        type=str,
        default=environ.get("DBSERVER", "127.0.0.1"),
        help="database host",
    )
    parser.add_argument(
        "--dbport",
        type=str,
        default=environ.get("DBPORT", "3306"),
        help="database port",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=int(environ.get("PORT", 1883)),
        help="port to listen on",
    )
    parser.add_argument(
        "-b",
        "--bind",
        type=str,
        default=environ.get("BIND", ""),
        help="address to bind to",
    )
    parser.add_argument(
        "-x", "--ping", action="store_true", help="ping database end exit"
    )
    return parser.parse_args(arguments)


if __name__ == "__main__":
    args = get_args()
    db = MeasurementDatabase(
        args.database, args.dbhost, args.dbport, args.dbuser, args.dbpassword
    )

    if args.ping:
        exit()

    print(
        f"starting server, listening on {args.bind}:{args.port}",
        file=stderr,
        flush=True,
    )
    while True:  # apparently serve_forever() does return on a 104 error
        server = Interceptor((args.bind, args.port), db)
        server.serve_forever()
        print("restarting server on a 104 error", file=stderr, flush=True)



import pytest
from io import BytesIO as IO
from unittest import mock
from datetime import datetime, timedelta

from htcollector.Server import InterceptorHandlerFactory
from htcollector.Database import MeasurementDatabase, Measurement


@pytest.fixture(scope="class")
def database():
    db = MeasurementDatabase(
        database="shellyht",
        host="127.0.0.1",
        port="3306",
        user="test-user",
        password="test_secret",
    )
    return db


# helper class and functions to mock the activity of the
# InterceptorHandler instance we will get from the factory
class MockSocket(object):
    def getsockname(self):
        return ("sockname",)


class MockRequest(object):
    _sock = MockSocket()

    def __init__(self, path):
        self._path = path

    def makefile(self, *args, **kwargs):
        if args[0] == "rb":
            return IO(b"GET %s HTTP/1.0" % self._path)
        elif args[0] == "wb":
            return IO(b"")
        else:
            raise ValueError("Unknown file type to make", args, kwargs)


def finish(self):
    # Do not close self.wfile, so we can read its value
    self.wfile.flush()
    self.rfile.close()


def date_time_string(self, timestamp=None):
    """ Mocked date time string """
    return "DATETIME"


def version_string(self):
    """ mock the server id """
    return "BaseHTTP/x.x Python/x.x.x"


class TestInterceptor:
    def test_GET_OK(self, database, capsys):
        stationid = "testid-123456"
        interceptorhandler = InterceptorHandlerFactory.getHandler(database)

        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/sensorlog?hum=70&temp=24&id=%s"
                            % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert (
                            ihinstance.wfile.getvalue()
                            == b"HTTP/1.0 200 OK\r\nServer: BaseHTTP/x.x Python/x.x.x\r\nDate: DATETIME\r\n\r\n"
                        )
                        database.names(
                            stationid, "testroom1"
                        )  # without a stationid mapping we never get anything back
                        r = database.retrieveMeasurements(stationid, start)
                        print(start, r)
                        assert len(r) == 1

    def test_GET_FORBIDDEN(self, database, capsys):
        stationid = "testid-@123456"  # contains an illegal character
        interceptorhandler = InterceptorHandlerFactory.getHandler(database)

        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/sensorlog?hum=70&temp=24&id=%s"
                            % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        response = ihinstance.wfile.getvalue()
                        print(response)
                        assert (
                            response
                            == b"HTTP/1.0 403 Forbidden\r\nServer: BaseHTTP/x.x Python/x.x.x\r\nDate: DATETIME\r\n\r\n"
                        )
                        database.names(
                            stationid, "testroom1"
                        )  # without a stationid mapping we never get anything back
                        r = database.retrieveMeasurements(stationid, start)
                        print(start, r)
                        assert len(r) == 0

    def test_GET_SERVER_ERROR(self, database, capsys):
        stationid = "testid-654321"
        interceptorhandler = InterceptorHandlerFactory.getHandler(
            None
        )  # no database reference will force an exception

        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/sensorlog?hum=70&temp=24&id=%s"
                            % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        response = ihinstance.wfile.getvalue()
                        print(response)
                        assert (
                            response
                            == b"HTTP/1.0 500 Internal Server Error\r\nServer: BaseHTTP/x.x Python/x.x.x\r\nDate: DATETIME\r\n\r\n"
                        )
                        database.names(
                            stationid, "testroom1"
                        )  # without a stationid mapping we never get anything back
                        r = database.retrieveMeasurements(stationid, start)
                        print(start, r)
                        assert len(r) == 0

    def test_GET_GRAPH(self, database, capsys):
        stationid = "graphid-123456"
        interceptorhandler = InterceptorHandlerFactory.getHandler(database)

        database.names(
            stationid, "testroom1"
        )  # without a stationid mapping we never get anything back
        start = datetime.now()
        database.storeMeasurement(Measurement(stationid, 10, 40))
        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/graph?id=%s" % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert ihinstance.wfile.getvalue()[:15] == b"HTTP/1.0 200 OK"

    def test_GET_HTML(self, database, capsys):
        stationid = "htmlid-123456"
        interceptorhandler = InterceptorHandlerFactory.getHandler(database)

        database.names(
            stationid, "htmlroom1"
        )  # without a stationid mapping we never get anything back
        start = datetime.now()
        database.storeMeasurement(Measurement(stationid, 10, 40))
        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/html?id=%s" % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert ihinstance.wfile.getvalue()[:15] == b"HTTP/1.0 200 OK"

    def test_GET_JSON(self, database, capsys):
        stationid = "jsonid-123456"
        interceptorhandler = InterceptorHandlerFactory.getHandler(database)

        database.names(
            stationid, "json room1"
        )  # without a stationid mapping we never get anything back
        start = datetime.now()
        database.storeMeasurement(Measurement(stationid, 10, 40))
        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/json?id=%s" % bytes(stationid, "UTF-8")
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert ihinstance.wfile.getvalue()[:15] == b"HTTP/1.0 200 OK"

    def test_GET_NAME(self, database, capsys):
        stationid = "newname-123456"
        name = "NewName"

        interceptorhandler = InterceptorHandlerFactory.getHandler(database)
        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(
                            b"/name?id=%s&name=%s"
                            % (bytes(stationid, "UTF-8"), bytes(name, "UTF-8"))
                        )
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert ihinstance.wfile.getvalue()[:15] == b"HTTP/1.0 200 OK"

        interceptorhandler = InterceptorHandlerFactory.getHandler(database)
        with mock.patch.object(interceptorhandler, "finish", finish):
            with mock.patch.object(
                interceptorhandler, "date_time_string", date_time_string
            ):
                with mock.patch.object(
                    interceptorhandler, "version_string", version_string
                ):
                    with mock.patch.object(interceptorhandler, "wbufsize", lambda: 1):
                        start = datetime.now()
                        request = MockRequest(b"/names")
                        ihinstance = interceptorhandler(
                            request, ("127.0.0.1", 12345), "testserver.example.org"
                        )
                        captured = capsys.readouterr()
                        print(captured.out)
                        assert ihinstance.wfile.getvalue()[:15] == b"HTTP/1.0 200 OK"




import json
import pytest

from datetime import datetime, timedelta

from htcollector import Utils


class TestUtils:
    def test_DatetimeEncoder(self):
        d = datetime(2000, 10, 9, 12, 0, 0)
        t = timedelta(0.5)

        datetime_json = json.dumps(d, cls=Utils.DatetimeEncoder)
        assert datetime_json == '"2000-10-09T12:00:00"'

        timedelta_json = json.dumps(t, cls=Utils.DatetimeEncoder)
        assert timedelta_json == '"12:00:00"'

        str_json = json.dumps("hello", cls=Utils.DatetimeEncoder)
        assert str_json == '"hello"'

        with pytest.raises(TypeError):
            json.dumps(..., cls=Utils.DatetimeEncoder)



#  shellyhtcollector, a python module to process sensor readings from Shelly H&T devices
#
# (C) 2022 Michel Anders (varkenvarken)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  version: 20220805094810

import json
from datetime import datetime, timedelta


class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



from datetime import datetime
from time import sleep
import pytest
from pytest import approx

from htcollector import Database


@pytest.fixture(scope="class")
def database():
    db = Database.MeasurementDatabase(
        database="shellyht",
        host="127.0.0.1",
        port="3306",
        user="test-user",
        password="test_secret",
    )
    return db


@pytest.fixture(scope="class")
def measurement(stationid):
    return Database.Measurement(stationid, 30, 65)


@pytest.fixture(scope="class")
def stationid():
    return "test-123456"


class TestNames:
    def test_names_insert(self, database):
        assert database.names("abcdef", "testroom") == []

    def test_names_list(self, database):
        assert database.names("*", "") == [("abcdef", "testroom")]

    def test_names_replace(self, database):
        assert database.names("abcdef", "testroom") == []
        assert database.names("*", "") == [("abcdef", "testroom")]


class TestMeasurements:
    def test_Measurement_init(self, stationid):
        m = Database.Measurement(stationid, 11.2, "70.5")
        assert type(m) is Database.Measurement
        assert m.humidity == approx(70.5)
        assert m.temperature == approx(11.2)
        assert m.stationid == stationid

    def test_Measurement_exceptions(self):
        with pytest.raises(ValueError):
            m = Database.Measurement("shellyht-123fed", "~11.2", 70.5)
        with pytest.raises(ValueError):
            m = Database.Measurement("shellyht-123fed", 11.2, "70.5%")
        with pytest.raises(ValueError):
            m = Database.Measurement("@shed-123fed", 11.2, 70.5)

    def test_storeMeasurement(self, measurement, database, stationid):
        database.names(
            stationid, "testroom"
        )  # without a stationid mapping we never get anything back
        n = database.storeMeasurement(measurement)
        assert n == 1
        r = database.retrieveLastMeasurement(stationid)
        assert len(r) == 1
        m = r[0]
        assert m["stationid"] == stationid

    def test_storeMeasurement_multi(self, database):
        database.names(
            "test-100001", "testroom1"
        )  # without a stationid mapping we never get anything back
        database.names(
            "test-100002", "testroom2"
        )  # without a stationid mapping we never get anything back
        database.storeMeasurement(Database.Measurement("test-100001", 10, 40))
        database.storeMeasurement(Database.Measurement("test-100002", 15, 45))
        r = database.retrieveLastMeasurement("*")
        assert len(r) > 1
        m1 = [m for m in r if m["stationid"] == "test-100001"]
        assert len(m1) == 1
        m2 = [m for m in r if m["stationid"] == "test-100002"]
        assert len(m2) == 1

    def test_retrieveMeasurements(self, database):
        database.names(
            "test-100001", "testroom1"
        )  # without a stationid mapping we never get anything back
        database.names(
            "test-100002", "testroom2"
        )  # without a stationid mapping we never get anything back
        start = datetime.now()
        sleep(1)
        database.storeMeasurement(Database.Measurement("test-100001", 10, 40))
        database.storeMeasurement(Database.Measurement("test-100002", 15, 45))
        r = database.retrieveMeasurements("*", start)
        assert len(r) > 1
        m1 = [m for m in r if m["stationid"] == "test-100001"]
        assert len(m1) == 1
        m2 = [m for m in r if m["stationid"] == "test-100002"]
        assert len(m2) == 1
        r = database.retrieveMeasurements("test-100001", start)
        assert len(r) == 1
        m1 = [m for m in r if m["stationid"] == "test-100001"]
        assert len(m1) == 1





#  shellyhtcollector, a python module to process sensor readings from Shelly H&T devices
#
# (C) 2022 Michel Anders (varkenvarken)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  version: 20220815134637

import re
import mariadb
from datetime import datetime, timedelta
from dateutil import tz


class Measurement:
    """
    Represents a measurement of temperature and humidity by a station.

    Args:
        stationid (str): station identification. Must contain only 1 or more alphnumeric characters or hyphens
        temperature (float): _description_
        humidity (float): _description_

    Raises:
        ValueError: if the stationid argument contains illegal characters or temperature or humidity arguments are not compatible to floats

    """

    idchars = re.compile(r"^[a-z01-9-]+$", re.IGNORECASE)

    def __init__(self, stationid, temperature, humidity):
        if re.match(self.idchars, stationid):
            self.stationid = stationid
        else:
            raise ValueError("stationid argument contains illegal characters")
        try:
            self.temperature = float(temperature)
            self.humidity = float(humidity)
        except ValueError as e:
            e.args = (
                "temperature and humidity arguments most be floats or convertible to floats",
            )
            raise e

    def __repr__(self):
        return f'Measurement("{self.stationid}", {self.temperature}, {self.humidity})'


class MeasurementDatabase:
    """
    Implements a databases containing measurements and station descriptions.

    The backing database should be a MariaDB database server.

    Args:
        database (str): name of the database
        host (str): hostname or ip-address of teh database server
        port (str): port that the database server is listening on
        user (str): username of a user with access privileges to the database
        password (str): password of the user

    """

    def __init__(self, database, host, port, user, password):
        self.connection = mariadb.connect(
            user=user, password=password, host=host, database=database
        )
        self.connection.auto_reconnect = True

        # the timestamp is configured for millisecond resolution
        cursor = self.connection.cursor()
        cursor.execute(
            """CREATE TABLE IF NOT EXISTS Measurements(
            Timestamp DATETIME(3) DEFAULT CURRENT_TIMESTAMP,
            Stationid VARCHAR(100),
            Temperature REAL,
            Humidity REAL);"""
        )
        cursor.execute("""CREATE INDEX IF NOT EXISTS ts ON Measurements(Timestamp);""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS si ON Measurements(Stationid);""")
        cursor.execute(
            """CREATE TABLE IF NOT EXISTS StationidToName(
            Stationid  VARCHAR(100) NOT NULL PRIMARY KEY,
            Name TEXT NOT NULL);"""
        )

    def storeMeasurement(self, measurement):
        """
        Store a measurement into the database.

        Args:
            measurement (Measurement): the measurement

        Measurements do not contain timestamps, the are added automatically.
        """
        cursor = self.connection.cursor()
        cursor.execute(
            """INSERT INTO Measurements(Stationid, Temperature, Humidity)
                                    VALUES (?,?,?)""",
            (measurement.stationid, measurement.temperature, measurement.humidity),
        )
        self.connection.commit()
        n = cursor.rowcount
        cursor.close()
        return n

    def retrieveMeasurements(
        self, stationid, starttime: datetime, endtime: datetime = None
    ):
        """
        Get measurements inside a given timeframe.

        Args:
            stationid (str): stationid or asterisk '*'
            starttime (datetime): starttime of measurement period (inclusive)
            endtime (datetime, optional): endtime of measurement period (inclusive) or None for now. Defaults to None.

        Returns:
            list: of dict(timestamp:t, stationid:id, temperature:t, humidity:h)
        """
        # timestamps in MariaDB are stored in UTC
        endtime = (
            endtime.astimezone(tz.UTC)
            if endtime is not None
            else datetime.now(tz=tz.UTC)
        )
        starttime = starttime.astimezone(tz.UTC)
        starttime = starttime.replace(
            microsecond=(starttime.microsecond // 1000) * 1000
        )  # round down to millis
        if stationid == "*":
            cursor = self.connection.cursor()
            cursor.execute(
                f"""SELECT Timestamp, Stationid, Temperature, Humidity
                        FROM Measurements
                        WHERE Timestamp >= ? AND Timestamp <= ?""",
                (starttime, endtime),
            )
            rows = cursor.fetchall()
        else:
            print("timestamps", starttime, endtime)
            cursor = self.connection.cursor()
            cursor.execute(
                f"""SELECT Timestamp, Stationid, Temperature, Humidity
                        FROM Measurements
                        WHERE Stationid = ? AND Timestamp >= ? AND Timestamp <= ?""",
                (stationid, starttime, endtime),
            )
            rows = cursor.fetchall()

        # mariadb / mysql timestamps are in UTC but returned as 'naive' datetime objects
        rows = [
            {
                "timestamp": row[0].replace(tzinfo=tz.UTC).astimezone(tz.tzlocal()),
                "stationid": row[1],
                "temperature": row[2],
                "humidity": row[3],
            }
            for row in rows
        ]

        return rows

    def retrieveLastMeasurement(self, stationid, _names=None, _unique_stations=None):
        """
        Return the last measurement data for a station or all stations.

        Args:
            stationid (str): the stationid or an asterisk '*'

        Returns:
            list: a list of dict objects, one for each station
        """
        print(stationid, _names, _unique_stations)
        if _names is None:
            _names = {nm[0]: nm[1] for nm in self.names("*")}

        if _unique_stations is None:
            _unique_stations = self.uniqueStations()

        if stationid == "*":
            rows = []
            for unique_station in _unique_stations:
                rows.extend(
                    self.retrieveLastMeasurement(
                        unique_station, _names=_names, _unique_stations=_unique_stations
                    )
                )
        else:
            # get the data
            cursor = self.connection.cursor()
            cursor.execute(
                """SELECT Timestamp as 'Timestamp [timestamp]', Stationid, Temperature, Humidity
            FROM Measurements
            WHERE Stationid = ? ORDER BY timestamp DESC LIMIT 1;""",
                (stationid,),
            )
            rows = cursor.fetchall()
            # mariadb / mysql timestamps are in UTC but returned as 'naive' datetime objects
            rows = [
                {
                    "time": row[0].replace(tzinfo=tz.UTC),
                    "deltat": datetime.now() - row[0],
                    "stationid": row[1],
                    "name": _names.get(row[1], "unknown"),
                    "temperature": row[2],
                    "humidity": row[3],
                }
                for row in rows
            ]
        return rows

    def retrieveMeasurementsLast24Hours(self, stationid):
        ltz = tz.tzlocal()
        d = timedelta(hours=24)
        end = datetime.now(tz=ltz)
        start = end - d
        return self.retrieveMeasurements(stationid, start, end)

    def uniqueStations(self):
        conn = self.connection
        cursor = conn.cursor()
        cursor.execute("SELECT DISTINCT(Stationid) FROM Measurements")
        return [row[0] for row in cursor.fetchall()]

    def names(self, stationid, name=None):
        """
        Insert or replace a name for a stationid, or return a list of all stations._

        Args:
            stationid (str): the shellyht station id or an asterisk '*'
            name (str)): the name to associate with a stationid (ignored if stationid is '*')

        Returns:
            list: either an empty list when inserting or replacing or a list of tuples(stationid, name)
        """
        if stationid == "*":
            conn = self.connection
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM StationidToName")
            return [(row[0], row[1]) for row in cursor.fetchall()]
        else:
            conn = self.connection
            cursor = conn.cursor()
            cursor.execute(
                "REPLACE StationidToName(Stationid, Name) VALUES(?,?)",
                (stationid, name),
            )
            conn.commit()
            return []



from datetime import date, datetime

import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np
from matplotlib import cm
from matplotlib.colors import ListedColormap
from matplotlib.ticker import PercentFormatter

blue2red = ListedColormap(
    np.dstack((np.linspace(0, 1, 256), np.zeros(256), np.linspace(1, 0, 256)))[0]
)


def graph(
    db,
    fname,
    stationid,
    starttime,
    endtime,
    title="Indoor measurements",
    facecolor="#eedddd",
    tcolor="#d22c2b",
    hcolor="#2c2cd2",
    font="Amaranth",
    fontcolor="#d22c2b",
    **kwargs,
):
    """
    Generate a temeprature and humidity graph for a station.

    Args:
        db (MeasurementDatabase): the measurement database
        fname (str): filename or "-" for stdout
        stationid (str): the station id
        starttime (datetime): the start time
        endtime (datetime): the end time
        title (str, optional): the title of the graph. Defaults to "Indoor measurements".
        facecolor (str, optional): background color of the graph. Defaults to "#eedddd".
        tcolor (str, optional): color for line and label of the temperature. Defaults to "#d22c2b".
        hcolor (str, optional): color for line and label of the humidity. Defaults to "#2c2cd2".
        font (str, optional): font for labels and titles. Defaults to "Amaranth".
        fontcolor (str, optional): color of the labels and titles. Defaults to "#d22c2b".
    """
    data = db.retrieveMeasurements(stationid, starttime, endtime)
    # duplicate last entry and set time to endtime to extend last measurement
    data.append(data[-1])
    data[-1]["timestamp"] = endtime
    data = [
        (d["timestamp"], d["stationid"], d["temperature"], d["humidity"]) for d in data
    ]
    data = np.array(data)
    time = data[:, 0]
    temp = data[:, 2]
    hum = data[:, 3]
    fig = plt.figure(**kwargs)
    plt.title(
        f"{title} (now: {temp[-1]:.1f}°C / {hum[-1]:.1f}% )",
        fontfamily=font,
        color=fontcolor,
    )
    ax = fig.axes[0]
    ax.set_xlim(left=starttime, right=endtime)
    ax.set_ylabel("Temp (°C)", color=tcolor)
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
    ax.set_facecolor(facecolor)
    ax.plot(time, temp, color=tcolor)
    ax2 = ax.twinx()
    ax2.set_ylabel("Humidity (%)", color=hcolor)
    ax2.plot(time, hum, color=hcolor)
    ax2.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
    fig.autofmt_xdate()
    plt.savefig(fname)



import pytest

from htcollector.__main__ import get_args

help_msg = """usage: pytest [-h] [--database DATABASE] [--dbhost DBHOST] [--dbport DBPORT]
              [-p PORT] [-b BIND] [-x]

optional arguments:
  -h, --help            show this help message and exit
  --database DATABASE   database schema
  --dbhost DBHOST       database host
  --dbport DBPORT       database port
  -p PORT, --port PORT  port to listen on
  -b BIND, --bind BIND  address to bind to
  -x, --ping            ping database end exit
"""


class TestMain:
    def test_get_args_help(self, capsys):
        with pytest.raises(SystemExit):
            args = get_args(["--help"])
        captured = capsys.readouterr()
        out = captured.out
        #assert out == help_msg



#  shellyhtcollector, a python module to process sensor readings from Shelly H&T devices
#
# (C) 2022 Michel Anders (varkenvarken)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  version: 20220815140019

from json import dumps
import re
from io import BytesIO as IO
from datetime import datetime, timedelta
from dateutil import tz
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer

from .Database import Measurement
from .Graph import graph
from .Utils import DatetimeEncoder


class InterceptorHandlerFactory:
    """
    Provides a single handler that returns an InterceptorHandler(BaseHTTPRequestHandler)
    that writes measurements to the provided MeasurementDatabase.
    """

    @staticmethod
    def getHandler(db):
        class InterceptorHandler(BaseHTTPRequestHandler):
            querypattern = re.compile(
                r"^/sensorlog\?hum=(?P<humidity>\d+(\.\d+)?)\&temp=(?P<temperature>\d+(\.\d+)?)\&id=(?P<stationid>[a-z01-9-]+)$",
                re.IGNORECASE,
            )
            graphpattern = re.compile(
                r"^/graph\?id=(?P<stationid>[a-z01-9-]+)$",
                re.IGNORECASE,
            )
            htmlpattern = re.compile(
                r"^/html\?id=(?P<stationid>[a-z01-9-]+)$",
                re.IGNORECASE,
            )
            jsonpattern = re.compile(
                r"^/json\?id=(?P<stationid>[a-z01-9-]+)$",
                re.IGNORECASE,
            )
            namespattern = re.compile(
                r"^(/name\?id=(?P<stationid>[a-z01-9-]+)\&name=(?P<name>([a-z01-9-]|\s)+))|(/names)$",
                re.IGNORECASE,
            )

            def do_GET(self):
                print(self.path, flush=True)
                if m := re.match(self.querypattern, self.path):
                    try:
                        measurement = Measurement(
                            m.group("stationid"),
                            m.group("temperature"),
                            m.group("humidity"),
                        )
                        db.storeMeasurement(measurement)
                        self.send_response(HTTPStatus.OK)
                    except Exception as e:
                        print(e)
                        self.send_response(HTTPStatus.INTERNAL_SERVER_ERROR)
                elif m := re.match(self.graphpattern, self.path):
                    try:
                        now = datetime.now(tz=tz.tzlocal())
                        yesterday = now - timedelta(1)
                        f = IO(b"")
                        graph(db, f, m.group("stationid"), yesterday, now)
                        png = f.getvalue()
                        self.send_response(HTTPStatus.OK)
                        self.send_header("Content-type", "image/png")
                        self.send_header("Content-Length", str(len(png)))
                        self.end_headers()
                        self.wfile.write(png)
                        f.close()
                    except Exception as e:
                        print(e)
                        self.send_response(HTTPStatus.INTERNAL_SERVER_ERROR)
                elif m := re.match(self.htmlpattern, self.path):
                    try:
                        ms = db.retrieveLastMeasurement("*")
                        mdivs = "\n".join(
                            f"""
                            <div class="measurement{' late' if m['deltat'].total_seconds()>24*3600 else ''}" id="{m["stationid"]}">
                            <div class="station">{m["name"]}</div>
                            <div class="time" data-time="{m["time"]}"></div>
                            <div class="temp">{m["temperature"]:.1f}<span class="degrees">°C</span></div>
                            <div class="hum">{m["humidity"]:.0f}<span class=percent>%</span></div>
                            </div>"""
                            for m in ms
                        )
                        style = """
                        .body {width:100%; }
                        .measurements {width: 90%;}
                        .measurement {float:left; width:200px;}
                        .station {float:left; font-size:16pt;}
                        .time {float:left; font-size:12pt;}
                        .temp {float:left; clear:both; font-size:40pt;}
                        .hum {float:left; font-size:12pt;}
                        .late { background-color:red; }
                        """
                        html = bytes(f"""<html>
                        <head><meta charset="UTF-8"><meta http-equiv="refresh" content="300"><title>Temperatuur binnen</title>
                        <style>{style}</style>
                        </head>
                        <body>
                        <div class="measurements">
                        {mdivs}
                        </div>
                        </body>
                        <script>document.querySelectorAll("[data-time]").forEach(function(e){{d=new Date(e.getAttribute("data-time"));e.innerHTML=d.getHours() + ":" + d.getMinutes().toString().padStart(2, '0')}})</script>
                        </html>
                        """, "UTF-8")
                        self.send_response(HTTPStatus.OK)
                        self.send_header("Content-type", "text/html")
                        self.send_header("Content-Length", str(len(html)))
                        self.end_headers()
                        self.wfile.write(html)
                    except Exception as e:
                        print(e)
                        self.send_response(HTTPStatus.INTERNAL_SERVER_ERROR)
                elif m := re.match(self.jsonpattern, self.path):
                    try:
                        json = bytes(
                            dumps(db.retrieveLastMeasurement("*"), cls=DatetimeEncoder),
                            encoding="UTF-8",
                        )
                        self.send_response(HTTPStatus.OK)
                        self.send_header("Content-type", "application/json")
                        self.send_header("Content-Length", str(len(json)))
                        self.end_headers()
                        self.wfile.write(json)
                    except Exception as e:
                        print(e)
                        self.send_response(HTTPStatus.INTERNAL_SERVER_ERROR)
                elif m := re.match(self.namespattern, self.path):
                    try:
                        if not self.path.startswith("/names"):
                            stationid = m.group("stationid")
                            name = m.group("name")
                            result = db.names(stationid, name)
                        names = db.names("*")
                        names = "\n".join(
                            f"<tr><td>{n[0]}</td><td>{n[1]}</td></tr>" for n in names
                        )
                        self.send_response(HTTPStatus.OK)
                        html = f"""<html><body>
                        <table>{names}</table>
                        </body></html>
                        """
                        self.send_header("Content-type", "text/html")
                        self.send_header("Content-Length", str(len(html)))
                        self.end_headers()
                        self.wfile.write(bytes(html, "UTF-8"))
                    except Exception as e:
                        print(e)
                        self.send_response(HTTPStatus.INTERNAL_SERVER_ERROR)
                else:
                    self.send_response(HTTPStatus.FORBIDDEN)
                self.end_headers()

        return InterceptorHandler


class Interceptor(ThreadingHTTPServer):
    allow_reuse_address = True

    def __init__(self, server_address, db):
        super().__init__(server_address, InterceptorHandlerFactory.getHandler(db))



from setuptools import setup

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="htcollector",
    version="0.1.5",
    description="Tools to log updates from Shelly HT devices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://varkenvarken.github.io/shellyhtcollector2/",
    author="varkenvarken",
    author_email="test@example.com",
    license="GPLv3",
    packages=["htcollector"],
    python_requires=">=3.8",
    install_requires=["mariadb==1.1.4", "python-dateutil==2.8.2"],
    extras_require={"graph": ["numpy==1.23.1", "matplotlib==3.5.3"]},
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



import pytest

from time import sleep
from datetime import datetime, timedelta
from dateutil import tz
from pathlib import Path

from htcollector.Graph import graph
from htcollector.Database import MeasurementDatabase, Measurement


@pytest.fixture(scope="class")
def database():
    db = MeasurementDatabase(
        database="shellyht",
        host="127.0.0.1",
        port="3306",
        user="test-user",
        password="test_secret",
    )
    return db


class TestGraph:
    def test_graph(self, database, tmp_path, capsys):
        stationid = "test-101101"
        database.names(
            stationid, "testroom1"
        )  # without a stationid mapping we never get anything back
        database.storeMeasurement(Measurement(stationid, 10, 40))
        sleep(1)
        database.storeMeasurement(Measurement(stationid, 15, 45))

        now = datetime.now(tz=tz.tzlocal())
        fivesecondsago = now - timedelta(seconds=5)

        fname = tmp_path / "graph.png"
        graph(database, fname, stationid, fivesecondsago, now)
        captured = capsys.readouterr()
        print(captured.out)
        assert Path.is_file(fname)
